/* Search for (short) uniquely realisable mate in chess variants */
/* Alain Brobecker - 2008 august */
/* 2008.09.02 - removed BUG for test of mate at MaxPly when K is in check,
                it only was in initialisation of Boards with FS!!! */
/* 2008.08.28 - set variables in CopyPlyAToPlyB as global
                tested if opposite K attacks square in IsSquareAttacked
                cleaned up the code  */
/* 2008.08.26 - inserted a dir variable so that it's not computed everytime */
/* 2008.08.17 - set WKx,WKy,BKx,BKy as global variables
                added test of mate at MaxPly when K is in check, but it BUGS?! */
/* 2008.08.15 - added (simple) recognition of similar EndMoves */
/* 2008.08.14 - added move generation for Pawns (ep but no promotion) */
/* 2008.08.13 - added function for printing moves
                added mate/stalemate recognition
                added move generation for Kings, Rooks, Bishops, Queens */
/* 2008.08.12 - added move generation for kNights, with a macro */
/* 2008.05.23 - modified IsKingInCheck(i) to IsSquareAttackedBy(i,c) so that
                it can be used to check castling */
/* 2008.05.18 - added IsKingInCheck(i) */
/* 2008.05.17 - start+speed tests: dim1 arrays are faster than dim2 */
#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

#define MaxPly 9
#define MaxNbEndMoves 9999

#define FALSE 0
#define TRUE 1

/* c is for color (0=White, 1=Black), must be lowest bit. */
/* 2nd bit is set for pieces. And Q=B+R. */
#define ES 0  /* 000000 empty square */
#define FS 1  /* 000001 forbidden squares around the board */
#define WK 2  /* 00001c K */
#define BK 3
#define WP 6  /* 00011c P */
#define BP 7
#define WN 10 /* 00101c N */
#define BN 11
#define WB 18 /* 01001c B */
#define BB 19
#define WR 34 /* 10001c R */
#define BR 35
#define WQ 50 /* 11001c Q */
#define BQ 51

/* Every Board is made of 21+8+2+8+2+8+2+8+2+8+2+8+2+8+2+8+21=120 chars */
/* The 21s and 2s are used for forbidden squares (FS) around the board */
/* We could gain space by having 10+MaxPly*(11+8+2+...+2+8+11)+10 instead */

unsigned char Board[(MaxPly+2)*12*10];
unsigned int NbMates[MaxPly+2],NbStalemates[MaxPly+2];
unsigned int EndMoves[2*MaxNbEndMoves],NbEndMoves=0;
double NbGames[MaxPly+2];

/* Print 8*8 central part of Board[Ply], mostly for debugging */
void PrintPly(Ply) int Ply; {
  int i,x,y;
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i++]) {
        case ES: printf("."); break; case FS: printf("*"); break;
        case WK: printf("K"); break; case BK: printf("k"); break;
        case WP: printf("P"); break; case BP: printf("p"); break;
        case WN: printf("N"); break; case BN: printf("n"); break;
        case WB: printf("B"); break; case BB: printf("b"); break;
        case WR: printf("R"); break; case BR: printf("r"); break;
        case WQ: printf("Q"); break; case BQ: printf("q"); break;
        default:printf("?"); break; /* error */
      }
    }
    printf("\n");
    i+=2;
  }
}

/* same as above, but print it in FEN notation */
void PrintFEN(Ply) int Ply; {
  int i,x,y,nb;
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    nb=0;
    for(x=0;x<8;x++) {
      if(Board[i]==ES) {
        nb++;
      } else {
        if(nb!=0) { printf("%d",nb); nb=0; }
        switch(Board[i]) {
          case FS: printf("*"); break;
          case WK: printf("K"); break; case BK: printf("k"); break;
          case WP: printf("P"); break; case BP: printf("p"); break;
          case WN: printf("N"); break; case BN: printf("n"); break;
          case WB: printf("B"); break; case BB: printf("b"); break;
          case WR: printf("R"); break; case BR: printf("r"); break;
          case WQ: printf("Q"); break; case BQ: printf("q"); break;
          default:printf("?"); break; /* error */
        }
      }
      i++;
    }
    if(nb!=0) { printf("%d",nb); }
    if(y!=7) { printf("/"); }
    i+=2;
  }
}

void PrintMove(Move) unsigned int Move; {
  int t;
  t=(Move>>26) & 63;
/*!!!    for(i=0;i<y;i++) { printf("  "); }*/
  printf("%d.",(t/2)+1); if((t&1)==1) { printf(".. "); }
  t=(Move>>20) & 62; /* piece */
  switch(t) {
    case WK: printf("K"); break;
    case WP: printf("P"); break;
    case WN: printf("N"); break;
    case WB: printf("B"); break;
    case WR: printf("R"); break;
    case WQ: printf("Q"); break;
    default: printf("***error1***");
  }
  t=((Move>>14)&127) - 21; /* square */
  printf("%c%d",t%10+'a',8-t/10);
  t=(Move>>8) & 62; /* piece */
  if((t==0) && ((Move&3)<2)) {
    printf("-");
  } else {
    printf("x");
    switch(t) {
      case WK: printf("K"); break;
      case ES:case WP: printf("P"); break; /* also en passant capture */
      case WN: printf("N"); break;
      case WB: printf("B"); break;
      case WR: printf("R"); break;
      case WQ: printf("Q"); break;
      default: printf("***error2***");
    }
  }
  t=((Move>>2)&127) - 21; /* square */
  printf("%c%d",t%10+'a',8-t/10);
  switch(Move&3) {
    case 0: printf("#"); break;
    case 1: printf("="); break;
    case 2: printf("ep#"); break;
    case 3: printf("ep="); break;
  }
}


/* Insert move from square index to square index+displacement into list */
unsigned int IM_Ply,IM_Move,IM_t,IM_Found;
void InsertMove(Index,Displacement,Type) int Index,Displacement,Type; {
  IM_Ply=Index/120;
  IM_Move=IM_Ply<<26;                      /* ply (6 bits) */
  IM_Move+=(Board[Index]&254)<<(21-1);  /* piece (5 bits) */
  IM_Move+=(Index-120*IM_Ply)<<14;      /* square in [0;119] (7 bits) */
  IM_Move+=(Board[Index+Displacement]&254)<<(9-1); /* piece (5 bits) */
  IM_Move+=(Index+Displacement-120*IM_Ply)<<2; /* square (7 bits) */
  IM_Move+=Type; /* 0=mate, 1=stalemate, 2=ep mate, 3=ep stalemate (2 bits) */
  IM_Found=FALSE;
  IM_t=0;
  while((IM_t<NbEndMoves) && (IM_Found==FALSE)) {
    if(IM_Move==EndMoves[2*IM_t]) {
      IM_Found=TRUE;
      EndMoves[2*IM_t+1]++;
    }
    IM_t++;
  }
  if(IM_Found==FALSE) {
    PrintMove(IM_Move); printf(";"); PrintFEN(IM_Ply); printf("\n");
    EndMoves[2*NbEndMoves]=IM_Move;
    EndMoves[2*NbEndMoves+1]=1;
    NbEndMoves++;
    if(NbEndMoves>=MaxNbEndMoves) {
      printf("***error: More than %d EndMoves!\n",MaxNbEndMoves); exit(-1);
    }
  }
}


/* Copy Board[PlyA] to Board[PlyB] */
int CP_i,CP_j,CP_x,CP_y;
void CopyPlyAToPlyB(PlyA,PlyB) int PlyA,PlyB; {
  CP_j=120*PlyB+21; CP_i=120*PlyA+21;
  for(CP_y=0;CP_y<8;CP_y++) {
    for(CP_x=0;CP_x<8;CP_x++) { Board[CP_j++]=Board[CP_i++]; }
    CP_j+=2; CP_i+=2;
  }
}

/**************************************************************************/
/* Returns TRUE if square "index" is attacked by a "color" piece, else    */
/* it returns FALSE */

#define IsSquareAttackeByKingOrLeaper(dir,piece) \
  if(Board[index+dir]==(WK | color)) { \
    return(TRUE); \
  } else { \
    ISA_i=1; \
    while(Board[ISA_i]==ES) { ISA_i+=1; } \
    if(((ISA_i & 1)==1) && (Board[index+ISA_i*dir] & (piece+1))==(piece | color)) { return(TRUE); } \
  }

int ISA_i;
int IsSquareAttackedBy(index,color) int index,color; {
  /* attacked by opposite color N? */
  if(Board[index-8]==(WN | color)) { return(TRUE); }
  if(Board[index-12]==(WN | color)) { return(TRUE); }
  if(Board[index-19]==(WN | color)) { return(TRUE); }
  if(Board[index-21]==(WN | color)) { return(TRUE); }
  if(Board[index+21]==(WN | color)) { return(TRUE); }
  if(Board[index+19]==(WN | color)) { return(TRUE); }
  if(Board[index+12]==(WN | color)) { return(TRUE); }
  if(Board[index+8]==(WN | color)) { return(TRUE); }
  /* attacked by opposite color R or Q? */
  IsSquareAttackeByKingOrLeaper(1,WR)
  IsSquareAttackeByKingOrLeaper(10,WR)
  IsSquareAttackeByKingOrLeaper(-10,WR)
  IsSquareAttackeByKingOrLeaper(-1,WR)
  return(FALSE);
}

/**************************************************************************/
/* Returns:
 * -1 if K of previous player is still in check (previous move was illegal)
 *  1 if K of current player is mate
 *  2 if K of current player is stalemate 
 *  0 otherwise. */
   
#define NextDepth(d) \
  switch(RecursiveSearch(Ply+1)) { \
      case 0: /* !!! InsertMove(i,d,1); */ NbLegalMoves++; break; \
      case -1: break; \
      case 1: InsertMove(i,d,0); NbLegalMoves++; NbMates[Ply]++; break; \
      case 2: InsertMove(i,d,1); NbLegalMoves++; NbStalemates[Ply]++; break; \
  }

/* macro used to generate moves for K,N */
#define GenerateMove(Displacement,t) \
  t=Board[i+Displacement]; \
  if((t==ES) || ((t & (1+2))==OppositePlayer)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+Displacement+120]=Board[i]; \
    Board[i+120]=ES; \
    NextDepth(Displacement) \
  }
  
/* macro used for R,Q */
#define GenerateLeaperMove(dir,t) \
  t=1; \
  while(Board[i+t*dir]==ES) { \
    if((t & 1)==1) { \
      CopyPlyAToPlyB(Ply,Ply+1); \
      Board[i+t*dir+120]=Board[i]; \
      Board[i+120]=ES; \
      NextDepth(t) \
    } \
    t++; \
  } \
  if(((t & 1)==1) && ((Board[i+t*dir] & (1+2))==OppositePlayer)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+t*dir+120]=Board[i]; \
    Board[i+120]=ES; \
    NextDepth(t) \
  }  
   
int WKx,WKy,BKx,BKy;
int RecursiveSearch(Ply) int Ply; {
  int i,x,y,OppositePlayer,CurrentPlayer,NbLegalMoves,dir,t0,t1;
  
  /* search Ks' positions and check distance between both */
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i]) {
        case WK: WKx=x; WKy=y; t0=i; break; /* t0=index fo WK */
        case BK: BKx=x; BKy=y; t1=i; break; /* t1=index fo BK */
      }
      i++;
    }
    i+=2;
  }
  x=WKx-BKx; if(x<0) { x=-x; }
  y=WKy-BKy; if(y<0) { y=-y; }
  if(x+y==1) { return(-1); /*!!! printf("dist WK-BK: (%d;%d)\n",x,y);*/ }
  
  /* Set t0=index of current player's K, t1=index of previous player's K,
     then see if previous player's K is still in check (move was illegal). */
  CurrentPlayer=Ply&1;
  if(CurrentPlayer==1) { t1=t1-t0; t0=t0+t1; t1=t0-t1; }
  if(IsSquareAttackedBy(t1,CurrentPlayer)==TRUE) { return(-1); }
  
  /* Set t0 to TRUE if current player's K is in check. Then quit recursion 
     according to depth. At MaxPly we only process cases were K is in check. */
  if(Ply>MaxPly) { return(0); }
  OppositePlayer=1-CurrentPlayer;
  t0=IsSquareAttackedBy(t0,OppositePlayer);
  if((Ply==MaxPly) && (t0==FALSE)) { return(0); }

  /* We must process this Ply, so generate all possible moves. */
  dir=2*CurrentPlayer-1; /* direction of pawns */
  CurrentPlayer+=2;
  OppositePlayer+=2;
  NbLegalMoves=0;
  i=i-80; /* 120*Ply+21; */
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      /* Board[i] contains a piece of current's player color? */
      if((Board[i]&(2+1))==CurrentPlayer) {
        switch(Board[i]&254) {
          /**** generate moves for P ****/
          /*!!! PROMOTION: (blanc et y=1) ou (noir et y=6) */
          /*!!! il faut le tester dans single step + capture */
          case WP:
            /* single step only */
            if(Board[i+10*dir]==ES) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+10*dir+120]=Board[i];
              Board[i+120]=ES;
              NextDepth(10*dir)
            }
            break;
          /**** generate moves for K ****/
          /*!!! actualise castling rights */
          case WK:
            GenerateMove(-1,t1)
            GenerateMove(-10,t1)
            GenerateMove(+1,t1)
            GenerateMove(+10,t1)
            break;
          /**** generate moves for N ****/
          case WN:
            GenerateMove(-8,t1)
            GenerateMove(-12,t1)
            GenerateMove(-19,t1)
            GenerateMove(-21,t1)
            GenerateMove(+21,t1)
            GenerateMove(+19,t1)
            GenerateMove(+12,t1)
            GenerateMove(+8,t1)
            break;
          default: 
            /**** generate moves for R/Q ****/
            if((Board[i]&WR)==WR) {
              GenerateLeaperMove(1,t1)
              GenerateLeaperMove(10,t1)
              GenerateLeaperMove(-10,t1)
              GenerateLeaperMove(-1,t1)
            }
          break;
        }
      }
      i++;
    } /* next x */
    i+=2;
  } /* next y*/
  /*!!! castlings, will use IsSquareAttackedBy(i,c) */
  /*!!! printf("nb of legal moves:%d\n",NbLegalMoves);  */
/*!!! It is possible to make a simpler handling of castling if Ply<=8 by
      just checking if K and R are on their normal square, but this will
      fail if ply>8, eg 1.Nh3 b6 2.Rg1 Ba6 3.Rh1 Bxg2 4.Bxg2 ~ 5.oo */
  
  /* check if previous move mated, stalemated or nothing */  
  if(NbLegalMoves==0) {
    if(t0==TRUE) { return(1); } else { return(2); }
  } else {
    NbGames[Ply]+=(double) NbLegalMoves;
    return(0);
  }
}

/**************************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  int c,i,x,y;
  setbuf(stdout, NULL); /* no buffering for output! */
  
  /* initialise arrays */
  for(c=0;c<MaxPly+1;c++) {
    NbGames[c]=0.0;
    NbMates[c]=0;
    NbStalemates[c]=0;
  }
  
  /* initialise Board for Ply 0 */
  for(c=0;c<(MaxPly+2)*12*10;c++) { Board[c]=FS; } /* forbidden squares*/
  i=21;for(y=0;y<8;y++){for(x=0;x<8;x++){Board[i++]=ES;}i+=2;} /* empty squares */
  Board[21]=BR;Board[22]=BN;Board[23]=BB;Board[24]=BQ;
  Board[25]=BK;Board[26]=BB;Board[27]=BN;Board[28]=BR;
  for(x=0;x<8;x++) { Board[31+x]=BP; }
  for(x=0;x<8;x++) { Board[81+x]=WP; }  
  Board[91]=WR;Board[92]=WN;Board[93]=WB;Board[94]=WQ;
  Board[95]=WK;Board[96]=WB;Board[97]=WN;Board[98]=WR;
/*  PrintPly(0); printf("\n"); */
 
  /* go */
  RecursiveSearch(0);
  
  /* Print the number of games */
  printf("\nply;    NbGames   ; NbMates ; NbStalemates\n");
  for(c=0;c<MaxPly;c++) {
    if(c!=MaxPly-1) {
      printf("%3d;%14.0lf;%9d;%9d\n",c+1,NbGames[c],NbMates[c],NbStalemates[c]);
    } else {
      printf("%3d;%14.0lf;%9d; untested\n",c+1,NbGames[c],NbMates[c]);
    }  
  }
  
  /* sort (stupidly) the EndMoves according to their occurence */
/*  for(x=1;x<(NbEndMoves-1)*2+1;x+=2) {
    i=x;
    c=EndMoves[x];
    for(y=x+2;y<NbEndMoves*2+1;y+=2) {
      if(EndMoves[y]<c) { c=EndMoves[y]; i=y; }
    }
    EndMoves[i]=EndMoves[x];
    EndMoves[x]=c;
    i--;
    c=EndMoves[i];
    EndMoves[i]=EndMoves[c-1];
    EndMoves[c-1]=c;
  }
*/  
  /* Print the ending moves */
  printf("\nEnding moves;occurences\n");
  for(c=0;c<NbEndMoves;c++) {
   PrintMove(EndMoves[c*2]);
   printf(";%d\n",EndMoves[c*2+1]);
  }
  exit(0);
}
