/* 2008.10.18 - Removed useless Board[i+120]=ES in pawns movement. */

/* pull  chess */
/* surroundings of board (FS) enlarged for easy recognition of impossible pulls */
/* !!! castling to actualis */

#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

/*#define CASTLING CASTLING*/ /* remove if you don't want castling to be implemented */

#define MaxNbPlies 14 /* must be < 4 bits */
#define MaxNbEndMoves 9999

#define FALSE 0
#define TRUE 1

/* c is for color (0=White, 1=Black), must be lowest bit. */
/* 2nd bit is set for pieces. And Q=B+R. */
#define ES 0  /* 000000 empty square */
#define FS 1  /* 000001 forbidden squares around the board */
#define WK 2  /* 00001c K */
#define BK 3
#define WP 6  /* 00011c P */
#define BP 7
#define WN 10 /* 00101c N */
#define BN 11
#define WB 18 /* 01001c B */
#define BB 19
#define WR 34 /* 10001c R */
#define BR 35
#define WQ 50 /* 11001c Q */
#define BQ 51

/* Every Board is made of 7*15+7+8+7+8+7+8+7+8+7+8+7+8+7+8+7+8+7+7*15=344 chars */
/* The 21s and 7s are used for forbidden squares (FS) around the board */
unsigned char Board[(MaxNbPlies+2)*344];

unsigned int NbPlies;
unsigned int epSquare[MaxNbPlies+3]; /* square where an ep capture is possible or 0 */
unsigned int NbMates[MaxNbPlies+2],NbStalemates[MaxNbPlies+2];
unsigned int EndMove[MaxNbEndMoves],EndMoveOccurence[MaxNbEndMoves];
unsigned char EndMovePosition[MaxNbEndMoves][64];
unsigned int NbEndMoves=0;
unsigned int NbUniqueEndMovesAtPly[MaxNbPlies+2];
unsigned int NbEndMovesAtPly[MaxNbPlies+2];

double NbGames[MaxNbPlies+2];
#ifdef CASTLING
  unsigned char CastlingRights[MaxNbPlies+3]; /*6 bits:BQR,WQR,BKR,WKR,BK,WK, 1 when it has moved*/
#endif

/* Print 8*8 central part of Board[Ply], mostly for debugging */
void PrintPly(Ply) int Ply; {
  int i,x,y;
  i=344*Ply+7*15+7;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i++]) {
        case ES: printf("."); break; case FS: printf("*"); break;
        case WK: printf("K"); break; case BK: printf("k"); break;
        case WP: printf("P"); break; case BP: printf("p"); break;
        case WN: printf("N"); break; case BN: printf("n"); break;
        case WB: printf("B"); break; case BB: printf("b"); break;
        case WR: printf("R"); break; case BR: printf("r"); break;
        case WQ: printf("Q"); break; case BQ: printf("q"); break;
        default:printf("?"); break; /* error */
      }
    }
    printf("\n");
    i+=7;
  }
}

/* Print EndMovePosition[EndMoveNb] in FEN notation */
int PF_x,PF_y,PF_i,PF_nb;
void PrintFEN(EndMoveNb) int EndMoveNb; {
  PF_i=0;
  for(PF_y=0;PF_y<8;PF_y++) {
    PF_nb=0;
    for(PF_x=0;PF_x<8;PF_x++) {
      if(EndMovePosition[EndMoveNb][PF_i]==ES) {
        PF_nb++;
      } else {
        if(PF_nb!=0) { printf("%d",PF_nb); PF_nb=0; }
        switch(EndMovePosition[EndMoveNb][PF_i]) {
          case FS: printf("*"); break;
          case WK: printf("K"); break; case BK: printf("k"); break;
          case WP: printf("P"); break; case BP: printf("p"); break;
          case WN: printf("N"); break; case BN: printf("n"); break;
          case WB: printf("B"); break; case BB: printf("b"); break;
          case WR: printf("R"); break; case BR: printf("r"); break;
          case WQ: printf("Q"); break; case BQ: printf("q"); break;
          default:printf("?"); break; /* error */
        }
      }
      PF_i++;
    }
    if(PF_nb!=0) { printf("%d",PF_nb); }
    if(PF_y!=7) { printf("/"); }
  }
}

/* Same as above, but print it with chessmaster9000 fonts */
int PF_color;
void PrintChessMaster9000(EndMoveNb) int EndMoveNb; {
  PF_i=0;
  PF_color=0;
  for(PF_y=0;PF_y<8;PF_y++) {
    for(PF_x=0;PF_x<8;PF_x++) {
      switch(EndMovePosition[EndMoveNb][PF_i]) {
        case ES: printf("%c",143+PF_color); break;
        case FS: printf("%c",248); break;
        case WK: printf("%c",176+2*PF_color); break;
        case BK: printf("%c",179+PF_color); break;
        case WP: printf("%c",253+2*PF_color); break;
        case BP: printf("%c",159-28*PF_color); break;
        case WN: printf("%c",149+PF_color); break;
        case BN: printf("%c",151+PF_color); break;
        case WB: printf("%c",155+2*PF_color); break;
        case BB: printf("%c",158+27*PF_color); break;
        case WR: printf("%c",164+2*PF_color); break;
        case BR: printf("%c",167+PF_color); break;
        case WQ: printf("%c",170+2*PF_color); break;
        case BQ: printf("%c",173+2*PF_color); break;
        default:printf("?"); break; /* error */
      }
      PF_i++;
      PF_color=PF_color ^ 1;
    }
    PF_color=PF_color ^ 1;
    printf("\n");
  }
}


/* used to print the ending move in standard format */
void PrintMove(Move,Mask) unsigned int Move,Mask; {
  int t;
  t=(Move>>26) & 63;
/*!!!    for(i=0;i<y;i++) { printf("  "); }*/
  printf("%d.",(t/2)+1); if((t&1)==1) { printf(".. "); }
  t=(Move>>20) & 62; /* piece without color */
  switch(t) {
    case WK: printf("K"); break;
    case WP: printf("P"); break;
    case WN: printf("N"); break;
    case WB: printf("B"); break;
    case WR: printf("R"); break;
    case WQ: printf("Q"); break;
    default: printf("***error1***");
  }
  if((Mask & 8)!=0) {
    t=((Move>>14)&63); /* square */
    printf("%c%d",t%8+'a',8-t/8);
  }
  t=(Move>>8) & 62; /* piece without color */
  if(t==0) { /* or if((t==0) && ((Move&3)<2)) { */
    if((Mask & 4)!=0) { printf("-"); } else { printf("?"); }
  } else {
    if((Mask & 4)!=0) { printf("x"); } else { printf("?"); }
    if((Mask & 2)!=0) {
      switch(t) {
        case WK: printf("K"); break;
        case ES:case WP: printf("P"); break; /* also en passant capture */
        case WN: printf("N"); break;
        case WB: printf("B"); break;
        case WR: printf("R"); break;
        case WQ: printf("Q"); break;
        default: printf("***error2***");
      }
    }
  }
  if((Mask & 1)!=0) {
    t=((Move>>2)&63); /* square */
    printf("%c%d",t%8+'a',8-t/8);
  }
  switch(Move&3) {
    case 0: printf("#"); break;
    case 1: printf("="); break;
    case 2: printf("*#"); break; /* special move: promotion, ep, castling */
    case 3: printf("*="); break;
  }
}

/* Insert move from square index to square index+displacement into list */
unsigned int IM_Ply,IM_Move,IM_Found,IM_y,IM_x,IM_t,IM_t2;
void InsertMove(Index,Displacement,Type) int Index,Displacement,Type; {
  IM_Ply=Index/344;
  IM_Move=IM_Ply<<26;                   /* ply (6 bits) */
  IM_Move+=(Board[Index]&254)<<(21-1);  /* piece (5 bits) */
  IM_x=Index-344*IM_Ply-7*15-7;
  IM_Move+=((IM_x%15)+((int) IM_x/15)*8)<<14;      /* square in [0;63] (7 bits) */
  IM_Move+=(Board[Index+Displacement]&254)<<(9-1); /* piece (5 bits) */
  IM_x=Index+Displacement-344*IM_Ply-7*15-7;
  IM_Move+=((IM_x%15)+((int) IM_x/15)*8)<<2; /* square (7 bits) */
  IM_Move+=Type; /* 0=mate, 1=stalemate, 2=ep mate, 3=ep stalemate (2 bits) */
  IM_Found=FALSE;
  IM_t=0;
  while((IM_t<NbEndMoves) && (IM_Found==FALSE)) {
    if(IM_Move==EndMove[IM_t]) {
      IM_Found=TRUE;
      EndMoveOccurence[IM_t]++;
    }
    IM_t++;
  }
  if(IM_Found==FALSE) {
    /* insert position for this EndMove */
    IM_t=344*IM_Ply+7*15+7;
    IM_t2=0;
    for(IM_y=0;IM_y<8;IM_y++) {
      for(IM_x=0;IM_x<8;IM_x++) {
        EndMovePosition[NbEndMoves][IM_t2]=Board[IM_t];
        IM_t2++;
        IM_t++;
      }
      IM_t+=7;
    }
    /* print move + position */
/*    PrintMove(IM_Move,15); printf(";");
    PrintFEN(NbEndMoves); printf("\n");
    PrintChessMaster9000(NbEndMoves); printf("\n");
*/    
    /* insert move in list then increase NbEndMoves */
    EndMove[NbEndMoves]=IM_Move;
    EndMoveOccurence[NbEndMoves]=1;
    NbEndMoves++;
    if(NbEndMoves>=MaxNbEndMoves) {
      printf("***error: More than %d EndMove!\n",MaxNbEndMoves); exit(-1);
    }
  }
}


/* Copy Board[PlyA] to Board[PlyB] */
int CP_i,CP_j,CP_x,CP_y;
void CopyPlyAToPlyB(PlyA,PlyB) int PlyA,PlyB; {
  CP_j=344*PlyB+7*15+7; CP_i=344*PlyA+7*15+7;
  for(CP_y=0;CP_y<8;CP_y++) {
    for(CP_x=0;CP_x<8;CP_x++) { Board[CP_j++]=Board[CP_i++]; }
    CP_j+=7; CP_i+=7;
  }
#ifdef CASTLING
  CastlingRights[PlyB]=CastlingRights[PlyA]; /* copy castling rights */
#endif
}

/**************************************************************************/
/* Returns TRUE if square "index" is attacked by a "color" piece, else    */
/* it returns FALSE */

#define IsSquareAttackeByKingOrRider(dir,piece) \
  ISA_i=index+dir; \
  if(Board[ISA_i]==(WK | color)) { \
    return(TRUE); \
  } else { \
    while(Board[ISA_i]==ES) { ISA_i+=dir; } \
    if((Board[ISA_i] & (piece+1))==(piece | color)) { return(TRUE); } \
  }

int ISA_i;
int IsSquareAttackedBy(index,color) int index,color; {
  /* attacked by opposite color N? */
  if(Board[index-15+2]==(WN | color)) { return(TRUE); }
  if(Board[index-15-2]==(WN | color)) { return(TRUE); }
  if(Board[index-30+1]==(WN | color)) { return(TRUE); }
  if(Board[index-30-1]==(WN | color)) { return(TRUE); }
  if(Board[index+15+2]==(WN | color)) { return(TRUE); }
  if(Board[index+15-2]==(WN | color)) { return(TRUE); }
  if(Board[index+30+1]==(WN | color)) { return(TRUE); }
  if(Board[index+30-1]==(WN | color)) { return(TRUE); }
  /* attacked by opposite color P? */
  if(Board[index+15-1-(15-1)*2*color]==(WP | color)) { return(TRUE); }
  if(Board[index+15+1-(15+1)*2*color]==(WP | color)) { return(TRUE); }
  /* attacked by opposite color K, B or Q? */
  IsSquareAttackeByKingOrRider(15-1,WB)
  IsSquareAttackeByKingOrRider(15+1,WB)
  IsSquareAttackeByKingOrRider(-15-1,WB)
  IsSquareAttackeByKingOrRider(-15+1,WB)
  /* attacked by opposite color R or Q? */
  IsSquareAttackeByKingOrRider(1,WR)
  IsSquareAttackeByKingOrRider(15,WR)
  IsSquareAttackeByKingOrRider(-15,WR)
  IsSquareAttackeByKingOrRider(-1,WR)
  return(FALSE);
}

/**************************************************************************/
/* Returns:
 * -1 if K of previous player is in check (previous move was illegal)
 *  1 if K of current player is mate
 *  2 if K of current player is stalemate 
 *  0 otherwise. */

void Pull(position,Displacement) int position,Displacement; {
  int t;
  t=Board[position-Displacement];
  while((t!=ES) && (t!=FS)) {
    Board[position]=Board[position-Displacement];
    position=position-Displacement;
    Board[position]=ES;
    t=Board[position-Displacement];
  }
}

    
#define NextDepth(d,type) \
  switch(RecursiveSearch(Ply+1)) { \
      case 0: /* InsertMove(i,d,1);*/ NbLegalMoves++; break; \
      case -1: break; \
      case 1: InsertMove(i,d,type); NbLegalMoves++; NbMates[Ply]++; break; \
      case 2: InsertMove(i,d,type+1); NbLegalMoves++; NbStalemates[Ply]++; break; \
  }

/* macro used to generate moves for Royal Leaper */
#define GenerateMoveK(Displacement,t) \
  t=Board[i+Displacement]; \
  if((t==ES) || ((t & (1+2))==OppositePlayer)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+Displacement+344]=Board[i]; \
    Board[i+344]=ES; \
    Pull(i+344,Displacement); \
    NextDepth(Displacement,0) \
  }

/* macro used to generate moves for Leaper */
#define GenerateMoveLeaper(Displacement,t) \
  t=Board[i+Displacement]; \
  if((t==ES) || ((t & (1+2))==OppositePlayer)) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+Displacement+344]=Board[i]; \
    Board[i+344]=ES; \
    Pull(i+344,Displacement); \
    NextDepth(Displacement,0) \
  }
  
/* macro used to generate moves for Rider */
#define GenerateRiderMove(dir,t) \
  t=dir; \
  while(Board[i+t]==ES) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+t+344]=Board[i]; \
    Board[i+344]=ES; \
    Pull(i+344,t); \
    NextDepth(t,0) \
    t+=dir; \
  } \
  if((Board[i+t] & (1+2))==OppositePlayer) { \
    CopyPlyAToPlyB(Ply,Ply+1); \
    Board[i+t+344]=Board[i]; \
    Board[i+344]=ES; \
    Pull(i+344,t); \
    NextDepth(t,0) \
  }  
   
int WKx,WKy,BKx,BKy;
int RecursiveSearch(Ply) int Ply; {
  int i,x,y,OppositePlayer,CurrentPlayer,NbLegalMoves,dir,t0,t1;
#ifdef CASTLING  
  int OldCastlingRights;
#endif  
  
  /* search Ks' positions and check distance between both */
  i=344*Ply+7*15+7;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i]) {
        case WK: WKx=x; WKy=y; t0=i; break; /* t0=index fo WK */
        case BK: BKx=x; BKy=y; t1=i; break; /* t1=index fo BK */
      }
      i++;
    }
    i+=7;
  }
  x=WKx-BKx; if(x<0) { x=-x; }
  y=WKy-BKy; if(y<0) { y=-y; }
  if((x<2) && (y<2)) { return(-1); /*!!! printf("dist WK-BK: (%d;%d)\n",x,y);*/ }
  
  /* Set t0=index of current player's K, t1=index of previous player's K,
     then see if previous player's K is in check (move was illegal). */
  CurrentPlayer=Ply&1;
  if(CurrentPlayer==1) { t1=t1-t0; t0=t0+t1; t1=t0-t1; }
  if(IsSquareAttackedBy(t1,CurrentPlayer)==TRUE) { return(-1); }
  
  /* Set t0 to TRUE if current player's K is in check. Then quit recursion 
     according to depth. At NbPlies we only process cases were K is in check. */
  if(Ply>NbPlies) { return(0); }
  OppositePlayer=1-CurrentPlayer;
  t0=IsSquareAttackedBy(t0,OppositePlayer);
  if((Ply==NbPlies) && (t0==FALSE)) { return(0); }

  /* We must process this Ply, so generate all possible moves. */
  dir=2*CurrentPlayer-1; /* direction of pawns */
  CurrentPlayer+=2;
  OppositePlayer+=2;
  NbLegalMoves=0;
  i=i-8*15; /* 344*Ply+7*15+7; */
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      /* Board[i] contains a piece of current's player color? */
      if((Board[i]&(2+1))==CurrentPlayer) {
        switch(Board[i]&254) {
          /**** generate moves for P ****/
          case WP:
            /* single step (and maybe double step) */
            if(Board[i+15*dir]==ES) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+344]=ES;
              Pull(i+344,15*dir);
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+15*dir+344]=WN+CurrentPlayer-2; /* to N */
                NextDepth(15*dir,2)
                Board[i+15*dir+344]=WB+CurrentPlayer-2; /* to B */
                NextDepth(15*dir,2)
                Board[i+15*dir+344]=WR+CurrentPlayer-2; /* to R */
                NextDepth(15*dir,2)
                Board[i+15*dir+344]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth(15*dir,2)
              } else {
                /* no promotion */
                Board[i+15*dir+344]=Board[i];
                NextDepth(15*dir,0)
                /* double step? */
                if(  ((y==6) && (dir==-1) && (Board[i+2*15*dir]==ES))
                   ||((y==1) && (dir== 1) && (Board[i+2*15*dir]==ES)) ) {
                  CopyPlyAToPlyB(Ply,Ply+1);
                  Board[i+344]=ES;
                  Board[i+2*15*dir+344]=Board[i];
                  epSquare[Ply+1]=i+15*dir+344; /* set ep right */
/*!!! never used                 Pull(i+344,2*15*dir);*/
                  NextDepth(2*15*dir,0)
                  epSquare[Ply+1]=0; /* restore usual ep right */
                }
              }
            }
            /* captures by pawn */
            if((Board[i+(15-1)*dir] & (2+1))==OppositePlayer) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+344]=ES;
              Pull(i+344,(15-1)*dir);
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+(15-1)*dir+344]=WN+CurrentPlayer-2; /* to N */
                NextDepth((15-1)*dir,2)
                Board[i+(15-1)*dir+344]=WB+CurrentPlayer-2; /* to B */
                NextDepth((15-1)*dir,2)
                Board[i+(15-1)*dir+344]=WR+CurrentPlayer-2; /* to R */
                NextDepth((15-1)*dir,2)
                Board[i+(15-1)*dir+344]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth((15-1)*dir,2)
              } else {
                /* no promotion */
                Board[i+(15-1)*dir+344]=Board[i];
                NextDepth((15-1)*dir,0)
              }
            }
            if((Board[i+(15+1)*dir] & (2+1))==OppositePlayer) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+344]=ES;
              Pull(i+344,(15+1)*dir);
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+(15+1)*dir+344]=WN+CurrentPlayer-2; /* to N */
                NextDepth((15+1)*dir,2)
                Board[i+(15+1)*dir+344]=WB+CurrentPlayer-2; /* to B */
                NextDepth((15+1)*dir,2)
                Board[i+(15+1)*dir+344]=WR+CurrentPlayer-2; /* to R */
                NextDepth((15+1)*dir,2)
                Board[i+(15+1)*dir+344]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth((15+1)*dir,2)
              } else {
                /* no promotion */
                Board[i+(15+1)*dir+344]=Board[i];
                NextDepth((15+1)*dir,0)
              }
            }
            /* en passant captures */
            if(i+(15-1)*dir==epSquare[Ply]) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+(15-1)*dir+344]=Board[i];
              Board[i+(15-1-15)*dir+344]=ES;
              Board[i+344]=ES;
              Pull(i+344,(15-1)*dir);
              NextDepth((15-1)*dir,2)
            }
            if(i+(15+1)*dir==epSquare[Ply]) {
              CopyPlyAToPlyB(Ply,Ply+1);
              Board[i+(15+1)*dir+344]=Board[i];
              Board[i+(15+1-15)*dir+344]=ES;
              Board[i+344]=ES;
              Pull(i+344,(15+1)*dir);
              NextDepth((15+1)*dir,2)
            }
          break;
          /**** generate moves for K ****/
          case WK:
#ifdef CASTLING
            OldCastlingRights=CastlingRights[Ply];
            CastlingRights[Ply]=OldCastlingRights | (CurrentPlayer-2+1); /* set bit for BK or WK to 1 */
#endif
            GenerateMoveK(-1,t1)
            GenerateMoveK(-15-1,t1)
            GenerateMoveK(-15,t1)
            GenerateMoveK(-15+1,t1)
            GenerateMoveK(+1,t1)
            GenerateMoveK(+15+1,t1)
            GenerateMoveK(+15,t1)
            GenerateMoveK(+15-1,t1)
#ifdef CASTLING
            CastlingRights[Ply]=OldCastlingRights;
#endif
          break;
          /**** generate moves for N ****/
          case WN:
            GenerateMoveLeaper(-15-2,t1)
            GenerateMoveLeaper(-15+2,t1)
            GenerateMoveLeaper(-2*15-1,t1)
            GenerateMoveLeaper(-2*15+1,t1)
            GenerateMoveLeaper(+2*15+1,t1)
            GenerateMoveLeaper(+2*15-1,t1)
            GenerateMoveLeaper(+15+2,t1)
            GenerateMoveLeaper(+15-2,t1)
          break;
          default:
            /**** generate moves for B/Q ****/
            if((Board[i]&WB)==WB) {
              GenerateRiderMove(15-1,t1)
              GenerateRiderMove(15+1,t1)
              GenerateRiderMove(-15-1,t1)
              GenerateRiderMove(-15+1,t1)
            }
            /**** generate moves for R/Q ****/
            if((Board[i]&WR)==WR) {
#ifdef CASTLING
              /* actualise castling rights if R */
              if((Board[i]&254)==WR) {
                OldCastlingRights=CastlingRights[Ply];
                if((i%120)==91-70*(CurrentPlayer-2)) {
                  CastlingRights[Ply]=OldCastlingRights | ((CurrentPlayer-2+1)*16);
                }
                if((i%120)==98-70*(CurrentPlayer-2)) {
                  CastlingRights[Ply]=OldCastlingRights | ((CurrentPlayer-2+1)*4);
                }
              }
#endif
              GenerateRiderMove(1,t1)
              GenerateRiderMove(15,t1)
              GenerateRiderMove(-15,t1)
              GenerateRiderMove(-1,t1)
#ifdef CASTLING
              if((Board[i]&254)==WR) {
                CastlingRights[Ply]=OldCastlingRights;
              }
#endif
            }
          break;
        }
      }
      i++;
    } /* next x */
    i+=7;
  } /* next y */
  
  /**** generate castlings if possible ****/
  /* It is possible to make a simpler handling of castling if Ply<=8 by */
  /* just checking if K and R are on their normal square, but this will */
  /* fail if ply>8, eg 1.Nh3 b6 2.Rg1 Ba6 3.Rh1 Bxg2 4.Bxg2 ~ 5.oo      */
  
#ifdef CASTLING
  if((CastlingRights[Ply] & (5<<(CurrentPlayer-2)))==0) { /* kingside castling */
    i=120*Ply+95-70*(CurrentPlayer-2); /* points on K of CurrentPlayer */
    if((Board[i+1]==ES) && (Board[i+2]==ES)
       && (IsSquareAttackedBy(i,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i+1,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i+2,OppositePlayer-2)==FALSE)) {
      CopyPlyAToPlyB(Ply,Ply+1);
      Board[i+2+120]=Board[i]; Board[i+120]=ES; /* move K */
      Board[i+1+120]=Board[i+3]; Board[i+120+3]=ES; /* move R */
      NextDepth(2,2)
    }
  }
  if((CastlingRights[Ply] & (17<<(CurrentPlayer-2)))==0) { /* queenside castling */
    i=120*Ply+95-70*(CurrentPlayer-2);
    if((Board[i-1]==ES) && (Board[i-2]==ES) && (Board[i-3]==ES)
       && (IsSquareAttackedBy(i,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i-1,OppositePlayer-2)==FALSE)
       && (IsSquareAttackedBy(i-2,OppositePlayer-2)==FALSE)) {
      CopyPlyAToPlyB(Ply,Ply+1);
      Board[i-2+120]=Board[i]; Board[i+120]=ES; /* move K */
      Board[i-1+120]=Board[i-4]; Board[i+120-4]=ES; /* move R */
      NextDepth(-2,2)
    }
  }
#endif

  /* check if previous move mated, stalemated or nothing */  
  if(NbLegalMoves==0) {
    if(t0==TRUE) { return(1); } else { return(2); }
  } else {
    NbGames[Ply]+=(double) NbLegalMoves;
    return(0);
  }
}

/**************************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  int c,i,x,y;
  setbuf(stdout, NULL); /* no buffering for output! */
  
  if(argc<2) { printf("usage: %s NbPlies\n",argv[0]); exit(1); }
  c=argv[1][0];
  if((c<'1') || (c>'9')) { printf("usage: %s NbPlies\n",argv[0]); exit(1); }
  NbPlies=c-'0';
  c=argv[1][1];
  if((c>='0') && (c<='9')) { NbPlies=NbPlies*10+c-'0'; }
  
  /* initialise arrays */
#ifdef CASTLING
  CastlingRights[0]=0; /* copied from previous ply */
#endif  
  for(c=0;c<MaxNbPlies+2;c++) {
    epSquare[c]=0;      /* no ep capture possible */
    NbGames[c]=0.0;
    NbMates[c]=0;
    NbStalemates[c]=0;
    NbEndMovesAtPly[c]=0;
    NbUniqueEndMovesAtPly[c]=0;
  }
  epSquare[MaxNbPlies+2]=0; /*!!! useless? */

  /* put forbidden squares in all boards */  
  for(c=0;c<(MaxNbPlies+2)*344;c++) { Board[c]=FS; }
  /* initialise Board for Ply 0 */
  i=7*15+7;for(y=0;y<8;y++){for(x=0;x<8;x++){Board[i++]=ES;}i+=7;} /* empty squares */
  Board[112]=BR;Board[113]=BN;Board[114]=BB;Board[115]=BQ;
  Board[116]=BK;Board[117]=BB;Board[118]=BN;Board[119]=BR;
  for(x=0;x<8;x++) { Board[127+x]=BP; }
  for(x=0;x<8;x++) { Board[202+x]=WP; }  
  Board[217]=WR;Board[218]=WN;Board[219]=WB;Board[220]=WQ;
  Board[221]=WK;Board[222]=WB;Board[223]=WN;Board[224]=WR;

  /* go */
  RecursiveSearch(0);
  
  /* sort (stupidly) the EndMove according to their occurence */
  if(NbEndMoves!=0) {
    for(x=0;x<NbEndMoves-1;x++) {
      i=x;
      c=EndMoveOccurence[x];
      for(y=x+1;y<NbEndMoves;y++) {
        if(EndMoveOccurence[y]<c) { c=EndMoveOccurence[y]; i=y; }
      }
      c=EndMove[x];
      EndMove[x]=EndMove[i];
      EndMove[i]=c;
      c=EndMoveOccurence[x];
      EndMoveOccurence[x]=EndMoveOccurence[i];
      EndMoveOccurence[i]=c;
      for(y=0;y<64;y++) {
        c=EndMovePosition[x][y];
        EndMovePosition[x][y]=EndMovePosition[i][y];
        EndMovePosition[i][y]=c;
      }
    }
  }
  
  /* Print the ending moves */
  printf("\nEnding moves;occurences\n");
  for(i=0;i<NbPlies+1;i++) {
    for(c=0;c<NbEndMoves;c++) {
      x=EndMove[c];
      y=(x>>26);
      if(y==i) {
        NbEndMovesAtPly[y]++;
        if(EndMoveOccurence[c]==1) { NbUniqueEndMovesAtPly[y]++; }
        PrintMove(x,15);
        printf(";%d\n",EndMoveOccurence[c]);
      }
    }
  }
  
  /* Print the number of games */
  printf("\nply;    NbGames   ;NbEndMoves;NbUniqueEM; NbMates ; NbStalemates\n");
  for(c=0;c<NbPlies;c++) {
    if(c!=NbPlies-1) {
      printf("%3d;%14.0lf;%10d;%9d;%9d;%9d\n",c+1,NbGames[c],NbEndMovesAtPly[c],NbUniqueEndMovesAtPly[c],NbMates[c],NbStalemates[c]);
    } else {
      printf("%3d;%14.0lf;%10d;%9d;%9d; untested\n",c+1,NbGames[c],NbEndMovesAtPly[c],NbUniqueEndMovesAtPly[c],NbMates[c]);
    }  
  }
  
  /* search for minimal information for every unique EndMove */
  printf("\nPossible problems are:\n");
  unsigned int NbMinimal[8]; /* 8 types of minimal information, see below */
  c=0;
  while((EndMoveOccurence[c]==1) && (c<NbEndMoves)) {
    x=EndMove[c];
    for(i=0;i<8;i++) { NbMinimal[i]=0; }
    for(i=0;i<NbEndMoves;i++) {
      y=EndMove[i];
      /* Same ply, same playing piece, same type of endmove? */
      if( (x&((2047<<23)+3))==(y&((2047<<23)+3)) ) {
        NbMinimal[0]++; /* n.piece# */
        /* Same destination square? */
        if( (x&(127<<2))==(y&(127<<2)) ) {
          NbMinimal[4]++; /* n.piece?square# */
        }
        /* Both are moves without capture? */
        if( ((x&(31<<9))==0) && ((y&(31<<9))==0) ) {
          NbMinimal[1]++; /* n.piece-# */
          /* Same destination square? */
          if( (x&(127<<2))==(y&(127<<2)) ) {
            NbMinimal[5]++; /* n.piece-square# */
          }
        } else {
          /* Both are captures? */
          if( ((x&(31<<9))!=0) && ((y&(31<<9))!=0) ) {
            NbMinimal[2]++; /* n.piecex# */
            /* Same piece captured? */
            if( (x&(31<<9))==(y&(31<<9)) ) {
              NbMinimal[3]++; /* n.piecexpiece# */
              /* Also same destination square?*/
              if( (x&(127<<2))==(y&(127<<2)) ) {
                NbMinimal[7]++; /* n.piecexpiece_square# */
              }
            }
            /* Same destination square?*/
            if( (x&(127<<2))==(y&(127<<2)) ) {
              NbMinimal[6]++; /* n.piecexsquare# */
            }
          }
        }
      }
    }
    /* Print minimal info for this move */
    i=0;
    while((NbMinimal[i]!=1) && (i<8)) { i++; }
    switch(i) {
      case 0:printf("*"); PrintMove(x,0); break; /* n.piece# */
      case 1:                       /* n.piece-# */
      case 2:printf("*"); PrintMove(x,4); break; /* n.piecex# */
      case 3:printf("*"); PrintMove(x,6); break; /* n.piecexpiece# */
      case 4:printf("*"); PrintMove(x,1); break; /* n.piece?square# */
      case 5:                       /* n.piece-square# */
      case 6:printf("*"); PrintMove(x,5); break; /* n.piecexsquare# */
      case 7:printf("*"); PrintMove(x,7); break; /* n.piecexpiece_square# */
      default: PrintMove(x,15); break;
    }
/*!!! printf(" (");PrintMove(x,15); printf(") "); */
    printf("\n"); PrintFEN(c); printf("\n");
    PrintChessMaster9000(c); printf("\n");
    c++;
  }
  exit(0);
}
