//!!!Modify program so that the following doesn't need to go up to ply 10
//  1.a3
//  2.c6 3.Qc7
//  4.f3 5.h3 6.h4
//  7.Qg3#


// logical progressive chess (Paul Byway, Variant Chess 18, autumn 1995)
//  (1) White starts with one move; Black plays two consecutive moves, either with the same man or two different men; White then play three moves; and so on, the number of moves increasing by one each time the turn changes
//  (2) A player's turn ends if he gives check, regardless of how many moves he may have made.
//  (3) A player may not expose his own king to check at any time during his turn.
//  (4) A player whose king is in check must get out of check with the first move of his turn.
//  (5) A player who has no legal move or who runs out of legal moves during his turn is stalemated and the game is drawn.
//  (6) No pawn two advance and no castling (these rules were invented to speed up orthochess and are superfluous in the fast moving progressive chess).

/*           __                          Search for (short) uniquely
 _ ___ _____|  |__ _____    _ ___ ___ _  realisable mate in chess variants.
| '   |  _  |   __|  -__|  | '_  |  _' |
'_'_'_'___._'_____'_____'  |   __|___  | Alain Brobecker
                           '__'  '_____' 2008 august - 2009 september

2009.09.26 - Conditionnal compilation of pawn promotion.
             Baseline can be modified on input.
2009.02.07 - Modified sorting of endmoves.
2008.11.01 - Previous version was abusing of CopyPlyAToPlyB.
             Removed a bug at the end when NbEndMoves=0.
             Used dirx,diry instead of dir for pieces' movements.
2008.10.18 - Removed useless Board[i+120]=ES in pawns movement.
2008.10.12 - Added PrintChessMaster9000 but due to m$do$ converting
             chars the output only works with ">file" redirector.
2008.10.04 - Added NbPly as argument.
             Added array containing positions for EndMoves.
2008.09.21 - Added promotion.
             Conditionnal compilation of castling.
             Added a mask in PrintMove function.
             Counting nb of unique EndMoves for every ply.
2008.09.20 - Added castling.
2008.09.13 - Added recognition of minimal notation for a problem.
             Counting nb of EndMoves for every ply.
             Sorted EndMoves according to Ply and Occurences.
2008.09.02 - Removed BUG for test of mate at MaxNbPlies when K is in check,
             it only was in initialisation of Boards with FS!!!
2008.08.28 - Set variables in CopyPlyAToPlyB as global.
             Tested if opposite K attacks square in IsSquareAttacked (for castling).
             Cleaned up the code.
2008.08.26 - Inserted a dir variable so that it's not computed everytime.
2008.08.17 - Set WKx,WKy,BKx,BKy as global variables.
             Added test of mate at MaxNbPlies when K is in check, but it BUGS?!
2008.08.15 - Added (simple) recognition of similar EndMoves.
2008.08.14 - Added move generation for Pawns (ep but no promotion).
2008.08.13 - Added function for printing moves.
             Added mate/stalemate recognition.
             Added move generation for Kings, Rooks, Bishops, Queens.
2008.08.12 - Added move generation for kNights, with a macro.
2008.05.23 - Modified IsKingInCheck(i) to IsSquareAttackedBy(i,c) so that
             it can be used to check castling.
2008.05.18 - Added IsKingInCheck(i).
2008.05.17 - Start+speed tests: dim1 arrays are faster than dim2. */

#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

#define PROMOTION PROMOTION /* comment if you don't want pawn promotion to be implemented */

#define MaxNbPlies 30 /* must be < 6 bits */
#define MaxNbEndMoves 9999

#define FALSE 0
#define TRUE 1

/* c is for color (0=White, 1=Black), must be lowest bit. */
/* 2nd bit is set for pieces. And Q=B+R. */
#define ES 0  /* 000000 empty square */
#define FS 1  /* 000001 forbidden squares around the board */
#define WK 2  /* 00001c K */
#define BK 3
#define WP 6  /* 00011c P */
#define BP 7
#define WN 10 /* 00101c N */
#define BN 11
#define WB 18 /* 01001c B */
#define BB 19
#define WR 34 /* 10001c R */
#define BR 35
#define WQ 50 /* 11001c Q */
#define BQ 51

/* Every Board is made of 21+8+2+8+2+8+2+8+2+8+2+8+2+8+2+8+21=120 chars */
/* The 21s and 2s are used for forbidden squares (FS) around the board */
/* We could gain space by having 10+MaxNbPlies*(11+8+2+...+2+8+11)+10 instead */
unsigned char Board[(MaxNbPlies+2)*12*10];

unsigned int NbPlies,NbPliesToNextPlayer;
unsigned int NbMates[MaxNbPlies+2],NbStalemates[MaxNbPlies+2];
unsigned int EndMove[MaxNbEndMoves],EndMoveOccurence[MaxNbEndMoves];
unsigned char EndMovePosition[MaxNbEndMoves][64];
int NbEndMoves=0;
unsigned int NbUniqueEndMovesAtPly[MaxNbPlies+2];
unsigned int NbEndMovesAtPly[MaxNbPlies+2];

double NbGames[MaxNbPlies+2];

/* Print 8*8 central part of Board[Ply], mostly for debugging */
void PrintPly(Ply) int Ply; {
  int i,x,y;
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i++]) {
        case ES: printf("."); break; case FS: printf("*"); break;
        case WK: printf("K"); break; case BK: printf("k"); break;
        case WP: printf("P"); break; case BP: printf("p"); break;
        case WN: printf("N"); break; case BN: printf("n"); break;
        case WB: printf("B"); break; case BB: printf("b"); break;
        case WR: printf("R"); break; case BR: printf("r"); break;
        case WQ: printf("Q"); break; case BQ: printf("q"); break;
        default:printf("?"); break; /* error */
      }
    }
    printf("\n");
    i+=2;
  }
}

/* Print EndMovePosition[EndMoveNb] in FEN notation */
int PF_x,PF_y,PF_i,PF_nb;
void PrintFEN(EndMoveNb) int EndMoveNb; {
  PF_i=0;
  for(PF_y=0;PF_y<8;PF_y++) {
    PF_nb=0;
    for(PF_x=0;PF_x<8;PF_x++) {
      if(EndMovePosition[EndMoveNb][PF_i]==ES) {
        PF_nb++;
      } else {
        if(PF_nb!=0) { printf("%d",PF_nb); PF_nb=0; }
        switch(EndMovePosition[EndMoveNb][PF_i]) {
          case FS: printf("*"); break;
          case WK: printf("K"); break; case BK: printf("k"); break;
          case WP: printf("P"); break; case BP: printf("p"); break;
          case WN: printf("N"); break; case BN: printf("n"); break;
          case WB: printf("B"); break; case BB: printf("b"); break;
          case WR: printf("R"); break; case BR: printf("r"); break;
          case WQ: printf("Q"); break; case BQ: printf("q"); break;
          default:printf("?"); break; /* error */
        }
      }
      PF_i++;
    }
    if(PF_nb!=0) { printf("%d",PF_nb); }
    if(PF_y!=7) { printf("/"); }
  }
}

/* Same as above, but print it with chessmaster9000 fonts
   m$do$ output spoils it, use >out.txt instead. */
int PF_color;
void PrintChessMaster9000(EndMoveNb) int EndMoveNb; {
  PF_i=0;
  PF_color=0;
  for(PF_y=0;PF_y<8;PF_y++) {
    for(PF_x=0;PF_x<8;PF_x++) {
      switch(EndMovePosition[EndMoveNb][PF_i]) {
        case ES: printf("%c",143+PF_color); break;
        case FS: printf("%c",248); break;
        case WK: printf("%c",176+2*PF_color); break;
        case BK: printf("%c",179+PF_color); break;
        case WP: printf("%c",253+2*PF_color); break;
        case BP: printf("%c",159-28*PF_color); break;
        case WN: printf("%c",149+PF_color); break;
        case BN: printf("%c",151+PF_color); break;
        case WB: printf("%c",155+2*PF_color); break;
        case BB: printf("%c",158+27*PF_color); break;
        case WR: printf("%c",164+2*PF_color); break;
        case BR: printf("%c",167+PF_color); break;
        case WQ: printf("%c",170+2*PF_color); break;
        case BQ: printf("%c",173+2*PF_color); break;
        default:printf("?"); break; /* error */
      }
      PF_i++;
      PF_color=PF_color ^ 1;
    }
    PF_color=PF_color ^ 1;
    printf("\n");
  }
}

/* used to print the ending move in standard format */
void PrintMove(Move,Mask) unsigned int Move,Mask; {
  int t;
  t=(Move>>26) & 63;
/*!!!    for(i=0;i<y;i++) { printf("  "); }*/
//  switch(t) {
//    case 0: printf("1."); break;
//    case 1: printf("2."); break;
//    case 2: printf("2.?,"); break;
//    case 3: printf("3."); break;
//    case 4: printf("3.?,"); break;
//    case 5: printf("3.?,?,"); break;
//    case 6: printf("4."); break;
//    case 7: printf("4.?,"); break;
//    case 8: printf("4.?,?,"); break;
//    case 9: printf("4.?,?,?,"); break;
//    case 10: printf("5."); break;
//  }
  printf("%d.",t+1);


  t=(Move>>20) & 62; /* piece without color */
  switch(t) {
    case WK: printf("K"); break;
    case WP: printf("P"); break;
    case WN: printf("N"); break;
    case WB: printf("B"); break;
    case WR: printf("R"); break;
    case WQ: printf("Q"); break;
    default: printf("***error1***");
  }
  if((Mask & 8)!=0) {
    t=((Move>>14)&127) - 21; /* square */
    printf("%c%d",t%10+'a',8-t/10);
  }
  t=(Move>>8) & 62; /* piece without color */
  if(t==0) { /* or if((t==0) && ((Move&3)<2)) { */
    if((Mask & 4)!=0) { printf("-"); } else { printf("?"); }
  } else {
    if((Mask & 4)!=0) { printf("x"); } else { printf("?"); }
    if((Mask & 2)!=0) {
      switch(t) {
        case WK: printf("K"); break;
        case ES:case WP: printf("P"); break; /* also en passant capture */
        case WN: printf("N"); break;
        case WB: printf("B"); break;
        case WR: printf("R"); break;
        case WQ: printf("Q"); break;
        default: printf("***error2***");
      }
    }
  }
  if((Mask & 1)!=0) {
    t=((Move>>2)&127) - 21; /* square */
    printf("%c%d",t%10+'a',8-t/10);
  }
  switch(Move&3) {
    case 0: printf("#"); break;
    case 1: printf("="); break;
    case 2: printf("*#"); break; /* special move: promotion */
    case 3: printf("*="); break;
  }
}

/* Insert move from square index to square index+displacement into list */
unsigned int IM_Ply,IM_Move,IM_Found,IM_y,IM_x,IM_t,IM_t2;
void InsertMove(Index,Displacement,Type) int Index,Displacement,Type; {
  IM_Ply=Index/120;
  IM_Move=IM_Ply<<26;                   /* ply (6 bits) */
  IM_Move+=(Board[Index]&254)<<(21-1);  /* piece (5 bits) */
  IM_Move+=(Index-120*IM_Ply)<<14;      /* square in [0;119] (7 bits) */
  IM_Move+=(Board[Index+Displacement]&254)<<(9-1); /* piece (5 bits) */
  IM_Move+=(Index+Displacement-120*IM_Ply)<<2; /* square (7 bits) */
  IM_Move+=Type; /* 0=mate, 1=stalemate, 2=ep mate, 3=ep stalemate (2 bits) */
  IM_Found=FALSE;
  IM_t=0;
  while((IM_t<NbEndMoves) && (IM_Found==FALSE)) {
    if(IM_Move==EndMove[IM_t]) {
      IM_Found=TRUE;
      EndMoveOccurence[IM_t]++;
    }
    IM_t++;
  }
  if(IM_Found==FALSE) {
    /* insert position for this EndMove */
    IM_t=120*IM_Ply+21;
    IM_t2=0;
    for(IM_y=0;IM_y<8;IM_y++) {
      for(IM_x=0;IM_x<8;IM_x++) {
        EndMovePosition[NbEndMoves][IM_t2]=Board[IM_t];
        IM_t2++;
        IM_t++;
      }
      IM_t+=2;
    }
    /* print move + position */
/*    PrintMove(IM_Move,15); printf(";");
    PrintFEN(NbEndMoves); printf("\n");
    PrintChessMaster9000(NbEndMoves); printf("\n");
*/
    /* insert move in list then increase NbEndMoves */
    EndMove[NbEndMoves]=IM_Move;
    EndMoveOccurence[NbEndMoves]=1;
    NbEndMoves++;
    if(NbEndMoves>=MaxNbEndMoves) {
      printf("***error: More than %d EndMove!\n",MaxNbEndMoves); exit(-1);
    }
  }
}


/* Copy Board[PlyA] to Board[PlyB] */
int CP_i,CP_j,CP_x,CP_y;
void CopyPlyAToPlyB(PlyA,PlyB) int PlyA,PlyB; {
  CP_j=120*PlyB+21; CP_i=120*PlyA+21;
  for(CP_y=0;CP_y<8;CP_y++) {
    for(CP_x=0;CP_x<8;CP_x++) { Board[CP_j++]=Board[CP_i++]; }
    CP_j+=2; CP_i+=2;
  }
}

/**************************************************************************/
/* Returns TRUE if square "index" is attacked by a "color" piece, else    */
/* it returns FALSE */

#define IsSquareAttackedByKingOrRider(dirx,diry,piece) \
  ISA_i=index+diry*10+dirx; \
  if(Board[ISA_i]==(WK | color)) { \
    return(TRUE); \
  } else { \
    while(Board[ISA_i]==ES) { ISA_i+=diry*10+dirx; } \
    if((Board[ISA_i] & (piece+1))==(piece | color)) { return(TRUE); } \
  }

int ISA_i;
int IsSquareAttackedBy(index,color) int index,color; {
  /* attacked by opposite color N? */
  if(Board[index-8]==(WN | color)) { return(TRUE); }
  if(Board[index-12]==(WN | color)) { return(TRUE); }
  if(Board[index-19]==(WN | color)) { return(TRUE); }
  if(Board[index-21]==(WN | color)) { return(TRUE); }
  if(Board[index+21]==(WN | color)) { return(TRUE); }
  if(Board[index+19]==(WN | color)) { return(TRUE); }
  if(Board[index+12]==(WN | color)) { return(TRUE); }
  if(Board[index+8]==(WN | color)) { return(TRUE); }
  /* attacked by opposite color P? */
  if(Board[index+9-18*color]==(WP | color)) { return(TRUE); }
  if(Board[index+11-22*color]==(WP | color)) { return(TRUE); }
  /* attacked by opposite color K, B or Q? */
  IsSquareAttackedByKingOrRider(-1, 1,WB)
  IsSquareAttackedByKingOrRider( 1, 1,WB)
  IsSquareAttackedByKingOrRider(-1,-1,WB)
  IsSquareAttackedByKingOrRider( 1,-1,WB)
  /* attacked by opposite color K, R or Q? */
  IsSquareAttackedByKingOrRider( 1, 0,WR)
  IsSquareAttackedByKingOrRider( 0, 1,WR)
  IsSquareAttackedByKingOrRider( 0,-1,WR)
  IsSquareAttackedByKingOrRider(-1, 0,WR)
  return(FALSE);
}

/**************************************************************************/
/* Returns:
 * -1 if K of previous player is in check (previous move was illegal)
 *  1 if K of current player is mate
 *  2 if K of current player is stalemate
 *  0 otherwise. */

#define NextDepth(d,type) \
  switch(RecursiveSearch(Ply+1)) { \
      case 0: /*!!! InsertMove(i,d,1);*/ NbLegalMoves++; break; \
      case -1: break; \
      case 1: InsertMove(i,d,type); NbLegalMoves++; NbMates[Ply]++; break; \
      case 2: InsertMove(i,d,type+1); NbLegalMoves++; NbStalemates[Ply]++; break; \
  }

/* macro used to generate moves for Royal Leaper */
#define GenerateMoveK(dirx,diry,t) \
  t=Board[i+dirx+10*diry]; \
  if((t==ES) || ((t & (1+2))==OppositePlayer)) { \
    Board[i+dirx+10*diry+120]=Board[i]; \
    NextDepth(dirx+10*diry,0) \
    Board[i+dirx+10*diry+120]=t; \
  }

/* macro used to generate moves for Leaper */
#define GenerateMoveLeaper(dirx,diry,t) \
  t=Board[i+dirx+10*diry]; \
  if((t==ES) || ((t & (1+2))==OppositePlayer)) { \
    Board[i+dirx+10*diry+120]=Board[i]; \
    NextDepth(dirx+10*diry,0) \
    Board[i+dirx+10*diry+120]=t; \
  }

/* macro used to generate moves for Rider */
#define GenerateMoveRider(dirx,diry,t) \
  t=dirx+10*diry; \
  while(Board[i+t]==ES) { \
    Board[i+t+120]=Board[i]; \
    NextDepth(t,0) \
    Board[i+t+120]=ES; \
    t+=dirx+10*diry; \
  } \
  if((Board[i+t] & (1+2))==OppositePlayer) { \
    Board[i+t+120]=Board[i]; \
    NextDepth(t,0) \
    Board[i+t+120]=Board[i+t]; \
  }

int WKx,WKy,BKx,BKy;
int RecursiveSearch(Ply) int Ply; {
  int i,x,y,OppositePlayer,CurrentPlayer,NbLegalMoves,dir,t0,t1;
  
  /* search Ks' positions and check distance between both */
  i=120*Ply+21;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      switch(Board[i]) {
        case WK: WKx=x; WKy=y; t0=i; break; /* t0=index fo WK */
        case BK: BKx=x; BKy=y; t1=i; break; /* t1=index fo BK */
      }
      i++;
    }
    i+=2;
  }
  x=WKx-BKx; if(x<0) { x=-x; }
  y=WKy-BKy; if(y<0) { y=-y; }
  if((x<2) && (y<2)) { return(-1); /*!!! printf("dist WK-BK: (%d;%d)\n",x,y);*/ }

  // LOGICAL PROGRESSIVE CHESS (LPC): SET CURRENT AND OPPOSITE PLAYER
  //   0.w
  //   1.b 2.b
  //   3.w 4.w 5.w
  //   6.b 7.b 8.b 9.b
  //  10.w ...
  switch(Ply) {
    case 1: case 2: case 6: case 7: case 8: case 9: CurrentPlayer=1; break; //black
    default: CurrentPlayer=0; break; //white
  }
  OppositePlayer=1-CurrentPlayer;
    
  // Set t0=index of current player's K, t1=index of opposite player's K.
  // Then set t0/t1 to TRUE if current/opposite player's K is in check.
  if(CurrentPlayer==1) { t1=t1-t0; t0=t0+t1; t1=t0-t1; }
  t0=IsSquareAttackedBy(t0,OppositePlayer);
  t1=IsSquareAttackedBy(t1,CurrentPlayer);
  
  // SOME CHECKS ARE ILLEGAL OR FORBIDS A PLAYER TO FINISH HIS SERIE
  //  ply        0 1 2 3 4 5 6 7 8 9 a b
  //  who plays  W B B W W W B B B B W W
  //  current K+ I   I   I I   I I I   I  I:ILLEGAL
  //  oppos. K+  n I n I n n I n n n I n  n:no move
  
  // Current K in check and should not be?
  if((t0==TRUE) && (/*(Ply==0) ||*/ (Ply==2) || (Ply==4) || (Ply==5) || (Ply==7) || (Ply==8) || (Ply==9) || (Ply==11)) ) {
    return(-1);
  }  
  // Opposite K in check and should not be    
  if((t1==TRUE) && ((Ply==1) || (Ply==3) || (Ply==6) || (Ply==10)) ) {
    return(-1);
  }
  
  CopyPlyAToPlyB(Ply,Ply+1);
  
  if((t1==TRUE) && (/*(Ply==0) ||*/ (Ply==2) || (Ply==4) || (Ply==5) || (Ply==7) || (Ply==8) || (Ply==9) || (Ply==11)) ) {
  // No move
    return(RecursiveSearch(Ply+1));
    
  } else {

  if((Ply==NbPlies) && (t0==FALSE)) { return(0); }
  if(Ply>NbPliesToNextPlayer) { return(0); }
  
  /* We must process this Ply, so generate all possible moves. */
  dir=2*CurrentPlayer-1; /* direction of pawns */
  CurrentPlayer+=2;
  OppositePlayer+=2;
  NbLegalMoves=0;
  i=i-80; /* 120*Ply+21; */

  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      /* Board[i] contains a piece of current's player color? */
      if((Board[i]&(2+1))==CurrentPlayer) {
        Board[i+120]=ES;
        switch(Board[i]&254) {
          /**** generate moves for P ****/
          case WP:
            /* single step (and maybe double step) */
            if(Board[i+10*dir]==ES) {
#ifdef PROMOTION
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+10*dir+120]=WN+CurrentPlayer-2; /* to N */
                NextDepth(10*dir,2)
                Board[i+10*dir+120]=WB+CurrentPlayer-2; /* to B */
                NextDepth(10*dir,2)
                Board[i+10*dir+120]=WR+CurrentPlayer-2; /* to R */
                NextDepth(10*dir,2)
                Board[i+10*dir+120]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth(10*dir,2)
                Board[i+10*dir+120]=ES;
              } else {
#endif
                /* no promotion */
                Board[i+10*dir+120]=Board[i];
                NextDepth(10*dir,0)
                Board[i+10*dir+120]=ES;
#ifdef PROMOTION
              }
#endif
            }
            /* captures by pawn */
            if((Board[i+9*dir] & (2+1))==OppositePlayer) {
#ifdef PROMOTION
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+9*dir+120]=WN+CurrentPlayer-2; /* to N */
                NextDepth(9*dir,2)
                Board[i+9*dir+120]=WB+CurrentPlayer-2; /* to B */
                NextDepth(9*dir,2)
                Board[i+9*dir+120]=WR+CurrentPlayer-2; /* to R */
                NextDepth(9*dir,2)
                Board[i+9*dir+120]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth(9*dir,2)
                Board[i+9*dir+120]=Board[i+9*dir];
              } else {
#endif
                /* no promotion */
                Board[i+9*dir+120]=Board[i];
                NextDepth(9*dir,0)
                Board[i+9*dir+120]=Board[i+9*dir];
#ifdef PROMOTION
              }
#endif
            }
            if((Board[i+11*dir] & (2+1))==OppositePlayer) {
#ifdef PROMOTION
              if(((y==6) && (dir==1)) || ((y==1) && (dir==-1))) {
                /* promotion */
                Board[i+11*dir+120]=WN+CurrentPlayer-2; /* to N */
                NextDepth(11*dir,2)
                Board[i+11*dir+120]=WB+CurrentPlayer-2; /* to B */
                NextDepth(11*dir,2)
                Board[i+11*dir+120]=WR+CurrentPlayer-2; /* to R */
                NextDepth(11*dir,2)
                Board[i+11*dir+120]=WQ+CurrentPlayer-2; /* to Q */
                NextDepth(11*dir,2)
                Board[i+11*dir+120]=Board[i+11*dir];
              } else {
#endif
                /* no promotion */
                Board[i+11*dir+120]=Board[i];
                NextDepth(11*dir,0)
                Board[i+11*dir+120]=Board[i+11*dir];
#ifdef PROMOTION
              }
#endif
            }
          break;
          /**** generate moves for K ****/
          case WK:
            GenerateMoveK(-1, 0,t1)
            GenerateMoveK(-1,-1,t1)
            GenerateMoveK( 0,-1,t1)
            GenerateMoveK( 1,-1,t1)
            GenerateMoveK( 1, 0,t1)
            GenerateMoveK( 1, 1,t1)
            GenerateMoveK( 0, 1,t1)
            GenerateMoveK(-1, 1,t1)
          break;
          /**** generate moves for N ****/
          case WN:
            GenerateMoveLeaper( 2,-1,t1)
            GenerateMoveLeaper(-2,-1,t1)
            GenerateMoveLeaper( 1,-2,t1)
            GenerateMoveLeaper(-1,-2,t1)
            GenerateMoveLeaper(+1, 2,t1)
            GenerateMoveLeaper(-1, 2,t1)
            GenerateMoveLeaper( 2, 1,t1)
            GenerateMoveLeaper(-2, 1,t1)
          break;
          default:
            /**** generate moves for B/Q ****/
            if((Board[i]&WB)==WB) {
              GenerateMoveRider(-1, 1,t1)
              GenerateMoveRider( 1, 1,t1)
              GenerateMoveRider( 1,-1,t1)
              GenerateMoveRider(-1,-1,t1)
            }
            /**** generate moves for R/Q ****/
            if((Board[i]&WR)==WR) {
              GenerateMoveRider( 1, 0,t1)
              GenerateMoveRider( 0, 1,t1)
              GenerateMoveRider(-1, 0,t1)
              GenerateMoveRider( 0,-1,t1)
            }
          break;
        }
        Board[i+120]=Board[i];
      }
      i++;
    } /* next x */
    i+=2;
  } /* next y */

  /* check if previous move mated, stalemated or nothing */
  if(NbLegalMoves==0) {
    if(t0==TRUE) { return(1); } else { return(2); }
  } else {
    NbGames[Ply]+=(double) NbLegalMoves;
    return(0);
  }
  
  } //!!! go with "if((t1==TRUE) && ((Ply==0) ... { ... } else { ... "
  
}

/**************************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  int c,i,x,y;
  char Baseline[8]={'R','N','B','Q','K','B','N','R'};
  
  setbuf(stdout, NULL); /* no buffering for output! */

  if(argc<2) { printf("usage: %s NbPlies [RNBQKBNR]\n",argv[0]); exit(1); }
  c=argv[1][0];
  if((c<'1') || (c>'9')) { printf("usage: %s NbPlies [RNBQKBNR]\n",argv[0]); exit(1); }
  NbPlies=c-'0';
  c=argv[1][1];
  if((c>='0') && (c<='9')) { NbPlies=NbPlies*10+c-'0'; }
  if(argc==3) { for(c=0;c<8;c++) { Baseline[c]=argv[2][c]; } }
  
  switch(NbPlies) {
    case 1: NbPliesToNextPlayer=2; break;
    case 2:case 3: NbPliesToNextPlayer=4; break;
    case 4:case 5:case 6: NbPliesToNextPlayer=6; break;
    case 7:case 8:case 9:case 10: NbPliesToNextPlayer=11; break;
  }
  
  /* initialise arrays */
#ifdef CASTLING
  CastlingRights[0]=0; /* copied from previous ply */
#endif
  for(c=0;c<MaxNbPlies+2;c++) {
    NbGames[c]=0.0;
    NbMates[c]=0;
    NbStalemates[c]=0;
    NbEndMovesAtPly[c]=0;
    NbUniqueEndMovesAtPly[c]=0;
  }

  /* put forbidden squares in all boards */
  for(c=0;c<(MaxNbPlies+2)*12*10;c++) { Board[c]=FS; }
  
  /* initialise Board for Ply 0 */
  i=21;for(y=0;y<8;y++){for(x=0;x<8;x++){Board[i++]=ES;}i+=2;} /* empty squares */
  for(x=0;x<8;x++) { Board[31+x]=BP; Board[81+x]=WP; }
  for(x=0;x<8;x++) {
    switch(Baseline[x]) {
      case 'N': Board[21+x]=BN; Board[91+x]=WN; break;
      case 'B': Board[21+x]=BB; Board[91+x]=WB; break;
      case 'R': Board[21+x]=BR; Board[91+x]=WR; break;
      case 'Q': Board[21+x]=BQ; Board[91+x]=WQ; break;
      case 'K': Board[21+x]=BK; Board[91+x]=WK; break;
    }
  }

/*  Board[48]=WP; Board[88]=ES; *//* 1.h4 d5 2.h5 Nd7 3.h6 Ndf6 ... */
/*  Board[54]=BP; Board[34]=ES; *//* to test mate by 6.gxf8=N# */
/*  Board[46]=BN; Board[22]=ES; *//* Peter Rsler, Problemkiste 1994/08 */
//  PrintPly(0); printf("\n");


  CopyPlyAToPlyB(0,1);
//  Board[120+81]=ES; Board[120+71]=WP; // 1.a3
//  Board[120+82]=ES; Board[120+72]=WP; // 1.b3
//  Board[120+83]=ES; Board[120+73]=WP; // 1.c3
//  Board[120+84]=ES; Board[120+74]=WP; // 1.d3
//  Board[120+85]=ES; Board[120+75]=WP; // 1.e3
//  Board[120+86]=ES; Board[120+76]=WP; // 1.f3
//  Board[120+87]=ES; Board[120+77]=WP; // 1.g3
//  Board[120+88]=ES; Board[120+78]=WP; // 1.h3
//  Board[120+92]=ES; Board[120+71]=WN; // 1.Na3
//  Board[120+92]=ES; Board[120+73]=WN; // 1.Nc3
//  Board[120+97]=ES; Board[120+76]=WN; // 1.Nf3
  Board[120+97]=ES; Board[120+78]=WN; // 1.Nh3
//  PrintPly(1); printf("\n");
  
  /* go */
  RecursiveSearch(1);

  /* sort (stupidly) the EndMove according to their occurence */
  if(NbEndMoves!=0) {
    for(x=0;x<NbEndMoves-1;x++) {
      i=x;
      c=EndMoveOccurence[x];
      for(y=x+1;y<NbEndMoves;y++) {
        if(EndMoveOccurence[y]<c) { c=EndMoveOccurence[y]; i=y; }
      }
      c=EndMove[x];
      EndMove[x]=EndMove[i];
      EndMove[i]=c;
      c=EndMoveOccurence[x];
      EndMoveOccurence[x]=EndMoveOccurence[i];
      EndMoveOccurence[i]=c;
      for(y=0;y<64;y++) {
        c=EndMovePosition[x][y];
        EndMovePosition[x][y]=EndMovePosition[i][y];
        EndMovePosition[i][y]=c;
      }
    }
  }

  /* Print the ending moves */
  printf("\nEnding moves;occurences\n");
  for(i=0;i<NbPlies+1;i++) {
    for(c=0;c<NbEndMoves;c++) {
      x=EndMove[c];
      y=(x>>26);
      if(y==i) {
        NbEndMovesAtPly[y]++;
        if(EndMoveOccurence[c]==1) { NbUniqueEndMovesAtPly[y]++; }
        PrintMove(x,15);
        printf(";%d\n",EndMoveOccurence[c]);
      }
    }
  }

  /* Print the number of games */
  printf("\nply;    NbGames   ;NbEndMoves;NbUniqueEM; NbMates ; NbStalemates\n");
  for(c=0;c<NbPlies;c++) {
    if(c!=NbPlies-1) {
      printf("%3d;%14.0lf;%10d;%9d;%9d;%9d\n",c+1,NbGames[c],NbEndMovesAtPly[c],NbUniqueEndMovesAtPly[c],NbMates[c],NbStalemates[c]);
    } else {
      printf("%3d;%14.0lf;%10d;%9d;%9d; untested\n",c+1,NbGames[c],NbEndMovesAtPly[c],NbUniqueEndMovesAtPly[c],NbMates[c]);
    }
  }

  /* search for minimal information for every unique EndMove */
  printf("\nPossible problems are:\n");
  unsigned int NbMinimal[8]; /* 8 types of minimal information, see below */
  c=0;
  while((EndMoveOccurence[c]==1) && (c<NbEndMoves)) {
    x=EndMove[c];
    for(i=0;i<8;i++) { NbMinimal[i]=0; }
    for(i=0;i<NbEndMoves;i++) {
      y=EndMove[i];
      /* Same ply, same playing piece, same type of endmove? */
      if( (x&((2047<<21)+3))==(y&((2047<<21)+3)) ) {
        NbMinimal[0]++; /* n.piece# */
        /* Same destination square? */
        if( (x&(127<<2))==(y&(127<<2)) ) {
          NbMinimal[4]++; /* n.piece?square# */
        }
        /* Both are moves without capture? */
        if( ((x&(31<<9))==0) && ((y&(31<<9))==0) ) {
          NbMinimal[1]++; /* n.piece-# */
          /* Same destination square? */
          if( (x&(127<<2))==(y&(127<<2)) ) {
            NbMinimal[5]++; /* n.piece-square# */
          }
        } else {
          /* Both are captures? */
          if( ((x&(31<<9))!=0) && ((y&(31<<9))!=0) ) {
            NbMinimal[2]++; /* n.piecex# */
            /* Same piece captured? */
            if( (x&(31<<9))==(y&(31<<9)) ) {
              NbMinimal[3]++; /* n.piecexpiece# */
              /* Also same destination square?*/
              if( (x&(127<<2))==(y&(127<<2)) ) {
                NbMinimal[7]++; /* n.piecexpiece_square# */
              }
            }
            /* Same destination square?*/
            if( (x&(127<<2))==(y&(127<<2)) ) {
              NbMinimal[6]++; /* n.piecexsquare# */
            }
          }
        }
      }
    }
    /* Print minimal info for this move */
    i=0;
    while((NbMinimal[i]!=1) && (i<8)) { i++; }
    switch(i) {
      case 0:printf("*"); PrintMove(x,0); break; /* n.piece# */
      case 1:                       /* n.piece-# */
      case 2:printf("*"); PrintMove(x,4); break; /* n.piecex# */
      case 3:printf("*"); PrintMove(x,6); break; /* n.piecexpiece# */
      case 4:printf("*"); PrintMove(x,1); break; /* n.piece?square# */
      case 5:                       /* n.piece-square# */
      case 6:printf("*"); PrintMove(x,5); break; /* n.piecexsquare# */
      case 7:printf("*"); PrintMove(x,7); break; /* n.piecexpiece_square# */
      default: PrintMove(x,15); break;
    }
/*!!! printf(" (");PrintMove(x,15); printf(") "); */
    printf("\n"); PrintFEN(c); printf("\n");
//    PrintChessMaster9000(c); printf("\n");
    c++;
  }
  exit(0);
}
